"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import numbers

import numpy as np
from sklearn.utils import check_array, resample

from .utils import find_all_paths, calculate_total_effect


class BootstrapMixin:
    """Mixin class for all LiNGAM algorithms that implement the method of bootstrapping."""

    def bootstrap(self, X, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        result : BootstrapResult
            Returns the result of bootstrapping.
        """
        # Check parameters
        X = check_array(X)

        if isinstance(n_sampling, (numbers.Integral, np.integer)):
            if not 0 < n_sampling:
                raise ValueError("n_sampling must be an integer greater than 0.")
        else:
            raise ValueError("n_sampling must be an integer greater than 0.")

        # Bootstrapping
        adjacency_matrices = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        total_effects = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        index = np.arange(X.shape[0])
        resampled_indices = []
        for i in range(n_sampling):
            resampled_X, resampled_index = resample(X, index)
            self.fit(resampled_X)
            adjacency_matrices[i] = self._adjacency_matrix

            # Calculate total effects
            for c, from_ in enumerate(self._causal_order):
                for to in self._causal_order[c + 1 :]:
                    total_effects[i, to, from_] = calculate_total_effect(
                        self._adjacency_matrix, from_, to
                    )

            resampled_indices.append(resampled_index)

        return BootstrapResult(adjacency_matrices, total_effects, resampled_indices=resampled_indices)


class BootstrapResult(object):
    """The result of bootstrapping."""

    def __init__(self, adjacency_matrices, total_effects, resampled_indices=None):
        """Construct a BootstrapResult.

        Parameters
        ----------
        adjacency_matrices : array-like, shape (n_sampling)
            The adjacency matrix list by bootstrapping.
        total_effects : array-like, shape (n_sampling)
            The total effects list by bootstrapping.
        resampled_indices :  array-like, shape (n_sampling, resample_size), optional (default=None)
            The list of original index of resampled samples.
        """
        self._adjacency_matrices = adjacency_matrices
        self._total_effects = total_effects
        self._resampled_indices = resampled_indices

    @property
    def adjacency_matrices_(self):
        """The adjacency matrix list by bootstrapping.

        Returns
        -------
        adjacency_matrices_ : array-like, shape (n_sampling)
            The adjacency matrix list, where ``n_sampling`` is
            the number of bootstrap sampling.
        """
        return self._adjacency_matrices

    @property
    def total_effects_(self):
        """The total effect list by bootstrapping.

        Returns
        -------
        total_effects_ : array-like, shape (n_sampling)
            The total effect list, where ``n_sampling`` is
            the number of bootstrap sampling.
        """
        return self._total_effects

    @property
    def resampled_indices_(self):
        """The list of original index of resampled samples.

        Returns
        -------
        resampled_indices_ : array-like, shape (n_sampling, resample_size)
            The list of original index of resampled samples,
            where ``n_sampling`` is the number of bootstrap sampling
            and ``resample_size`` is the size of each subsample set.
        """
        return self._resampled_indices

    def get_causal_direction_counts(
        self,
        n_directions=None,
        min_causal_effect=None,
        split_by_causal_effect_sign=False,
    ):
        """Get causal direction count as a result of bootstrapping.

        Parameters
        ----------
        n_directions : int, optional (default=None)
            If int, then The top ``n_directions`` items are included in the result
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects
            less than ``min_causal_effect`` are excluded.
        split_by_causal_effect_sign : boolean, optional (default=False)
            If True, then causal directions are split depending on the sign of the causal effect.

        Returns
        -------
        causal_direction_counts : dict
            List of causal directions sorted by count in descending order.
            The dictionary has the following format::

            {'from': [n_directions], 'to': [n_directions], 'count': [n_directions]}

            where ``n_directions`` is the number of causal directions.
        """
        # Check parameters
        if isinstance(n_directions, (numbers.Integral, np.integer)):
            if not 0 < n_directions:
                raise ValueError("n_directions must be an integer greater than 0")
        elif n_directions is None:
            pass
        else:
            raise ValueError("n_directions must be an integer greater than 0")

        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")

        # Count causal directions
        directions = []
        for am in np.nan_to_num(self._adjacency_matrices):
            direction = np.array(np.where(np.abs(am) > min_causal_effect))
            if split_by_causal_effect_sign:
                signs = (
                    np.array([np.sign(am[i][j]) for i, j in direction.T])
                    .astype("int64")
                    .T
                )
                direction = np.vstack([direction, signs])
            directions.append(direction.T)
        directions = np.concatenate(directions)

        if len(directions) == 0:
            cdc = {"from": [], "to": [], "count": []}
            if split_by_causal_effect_sign:
                cdc["sign"] = []
            return cdc

        directions, counts = np.unique(directions, axis=0, return_counts=True)
        sort_order = np.argsort(-counts)
        sort_order = (
            sort_order[:n_directions] if n_directions is not None else sort_order
        )
        counts = counts[sort_order]
        directions = directions[sort_order]

        cdc = {
            "from": directions[:, 1].tolist(),
            "to": directions[:, 0].tolist(),
            "count": counts.tolist(),
        }
        if split_by_causal_effect_sign:
            cdc["sign"] = directions[:, 2].tolist()

        return cdc

    def get_directed_acyclic_graph_counts(
        self, n_dags=None, min_causal_effect=None, split_by_causal_effect_sign=False
    ):
        """Get DAGs count as a result of bootstrapping.

        Parameters
        ----------
        n_dags : int, optional (default=None)
            If int, then The top ``n_dags`` items are included in the result
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than
            ``min_causal_effect`` are excluded.
        split_by_causal_effect_sign : boolean, optional (default=False)
            If True, then causal directions are split depending on the sign of the causal effect.

        Returns
        -------
        directed_acyclic_graph_counts : dict
            List of directed acyclic graphs sorted by count in descending order.
            The dictionary has the following format::

            {'dag': [n_dags], 'count': [n_dags]}.

            where ``n_dags`` is the number of directed acyclic graphs.
        """
        # Check parameters
        if isinstance(n_dags, (numbers.Integral, np.integer)):
            if not 0 < n_dags:
                raise ValueError("n_dags must be an integer greater than 0")
        elif n_dags is None:
            pass
        else:
            raise ValueError("n_dags must be an integer greater than 0")

        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")

        # Count directed acyclic graphs
        dags = []
        for am in np.nan_to_num(self._adjacency_matrices):
            dag = np.abs(am) > min_causal_effect
            if split_by_causal_effect_sign:
                direction = np.array(np.where(dag))
                signs = np.zeros_like(dag).astype("int64")
                for i, j in direction.T:
                    signs[i][j] = np.sign(am[i][j]).astype("int64")
                dag = signs
            dags.append(dag)

        dags, counts = np.unique(dags, axis=0, return_counts=True)
        sort_order = np.argsort(-counts)
        sort_order = sort_order[:n_dags] if n_dags is not None else sort_order
        counts = counts[sort_order]
        dags = dags[sort_order]

        if split_by_causal_effect_sign:
            dags = [
                {
                    "from": np.where(dag)[1].tolist(),
                    "to": np.where(dag)[0].tolist(),
                    "sign": [dag[i][j] for i, j in np.array(np.where(dag)).T],
                }
                for dag in dags
            ]
        else:
            dags = [
                {"from": np.where(dag)[1].tolist(), "to": np.where(dag)[0].tolist()}
                for dag in dags
            ]

        return {"dag": dags, "count": counts.tolist()}

    def get_probabilities(self, min_causal_effect=None):
        """Get bootstrap probability.

        Parameters
        ----------
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than
            ``min_causal_effect`` are excluded.

        Returns
        -------
        probabilities : array-like
            List of bootstrap probability matrix.
        """
        # check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")

        adjacency_matrices = np.nan_to_num(self._adjacency_matrices)
        shape = adjacency_matrices[0].shape
        bp = np.zeros(shape)
        for B in adjacency_matrices:
            bp += np.where(np.abs(B) > min_causal_effect, 1, 0)
        bp = bp / len(adjacency_matrices)

        if int(shape[1] / shape[0]) == 1:
            return bp
        else:
            return np.hsplit(bp, int(shape[1] / shape[0]))

    def get_total_causal_effects(self, min_causal_effect=None):
        """Get total effects list.

        Parameters
        ----------
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            If float, then causal directions with absolute values of causal effects less than
            ``min_causal_effect`` are excluded.

        Returns
        -------
        total_causal_effects : dict
            List of bootstrap total causal effect sorted by probability in descending order.
            The dictionary has the following format::

            {'from': [n_directions], 'to': [n_directions], 'effect': [n_directions], 'probability': [n_directions]}

            where ``n_directions`` is the number of causal directions.
        """
        # Check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")

        # Calculate probability
        probs = np.sum(
            np.where(np.abs(self._total_effects) > min_causal_effect, 1, 0),
            axis=0,
            keepdims=True,
        )[0]
        probs = probs / len(self._total_effects)

        # Causal directions
        dirs = np.array(np.where(np.abs(probs) > 0))
        probs = probs[dirs[0], dirs[1]]

        # Calculate median effect without zero
        effects = np.zeros(dirs.shape[1])
        for i, (to, from_) in enumerate(dirs.T):
            idx = np.where(np.abs(self._total_effects[:, to, from_]) > 0)
            effects[i] = np.median(self._total_effects[:, to, from_][idx])

        # Sort by probability
        order = np.argsort(-probs)
        dirs = dirs.T[order]
        effects = effects[order]
        probs = probs[order]

        ce = {
            "from": dirs[:, 1].tolist(),
            "to": dirs[:, 0].tolist(),
            "effect": effects.tolist(),
            "probability": probs.tolist(),
        }

        return ce

    def get_paths(self, from_index, to_index, min_causal_effect=None):
        """Get all paths from the start variable to the end variable and their bootstrap probabilities.

        Parameters
        ----------
        from_index : int
            Index of the variable at the start of the path.
        to_index : int
            Index of the variable at the end of the path.
        min_causal_effect : float, optional (default=None)
            Threshold for detecting causal direction.
            Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

        Returns
        -------
        paths : dict
            List of path and bootstrap probability.
            The dictionary has the following format::

            {'path': [n_paths], 'effect': [n_paths], 'probability': [n_paths]}

            where ``n_paths`` is the number of paths.
        """
        # check parameters
        if min_causal_effect is None:
            min_causal_effect = 0.0
        else:
            if not 0.0 < min_causal_effect:
                raise ValueError("min_causal_effect must be an value greater than 0.")

        # Find all paths from from_index to to_index
        paths_list = []
        effects_list = []
        for am in self._adjacency_matrices:
            paths, effects = find_all_paths(am, from_index, to_index, min_causal_effect)
            # Convert path to string to make them easier to handle.
            paths_list.extend(["_".join(map(str, p)) for p in paths])
            effects_list.extend(effects)

        paths_list = np.array(paths_list)
        effects_list = np.array(effects_list)

        # Count paths
        paths_str, counts = np.unique(paths_list, axis=0, return_counts=True)

        # Sort by count
        order = np.argsort(-counts)
        probs = counts[order] / len(self._adjacency_matrices)
        paths_str = paths_str[order]

        # Calculate median of causal effect for each path
        effects = [
            np.median(effects_list[np.where(paths_list == p)]) for p in paths_str
        ]

        result = {
            "path": [[int(i) for i in p.split("_")] for p in paths_str],
            "effect": effects,
            "probability": probs.tolist(),
        }
        return result



import numpy as np
import graphviz
import networkx as nx
from matplotlib.colors import is_color_like
from sklearn.utils import check_array, check_scalar


def get_common_edge_probabilities(bootstrap_results, mode='across'):
    """
    Calculate the probability of common edges across multiple bootstrap results.

    This function analyzes adjacency matrices from a list of bootstrap results and computes
    the probability of common edges (non-zero and not NaN) and hidden common cause edges (NaN)
    either per sample or across all samples depending on the selected mode.

    Parameters
    ----------
    bootstrap_results : list
        A list of objects, each containing an attribute `adjacency_matrices_`, which is a list
        or array of adjacency matrices (2D numpy arrays) from bootstrap iterations.

    mode : str, optional
        The mode of calculation. Must be either:

        * 'per': Computes common edge probabilities by extracting common edges at each bootstrap iteration.
        * 'across': Computes common edge probabilities across all samples within each bootstrap result.

    Returns
    -------
    common_edge_probabilities : np.ndarray
        A stacked array of shape (2, n, n), where:

        * common_edge_probabilities[0]: The probabilities of common edges.
        * common_edge_probabilities[1]: The nan probabilities of hidden common cause.
    """
    # check parameters
    if not isinstance(bootstrap_results, list):
        raise ValueError("bootstrap_results must be a list.")

    for bootstrap_result in bootstrap_results:
        if not hasattr(bootstrap_result, "adjacency_matrices_"):
            raise AttributeError("The bootstrap_result must implement a callable 'adjacency_matrices_'.")

    if mode == 'across':
        common_edge_prob_list = []
        hidden_common_edge_prob_list = []

        for bootstrap_result in bootstrap_results:
            matrices = np.array(bootstrap_result.adjacency_matrices_)

            # common edge: non-zero and not NaN
            valid_mask = ~np.isnan(matrices)
            nonzero_mask = (matrices != 0) & valid_mask
            common_edge_prob = nonzero_mask.sum(axis=0) / len(matrices)
            common_edge_prob_list.append(common_edge_prob)

            # hidden common cause edges: NaN
            unobserved_prob = np.isnan(matrices).sum(axis=0) / len(matrices)
            hidden_common_edge_prob_list.append(unobserved_prob)

        # Combine probabilities
        common_edge_prob = np.prod(common_edge_prob_list, axis=0)
        hidden_common_edge_prob = np.prod(hidden_common_edge_prob_list, axis=0)

        # Stack into a single array of shape (2, n, n)
        common_edge_probabilities = np.stack((common_edge_prob, hidden_common_edge_prob), axis=0)

    elif mode == 'per':
        n_samples = len(bootstrap_results[0].adjacency_matrices_)
        n_vars = bootstrap_results[0].adjacency_matrices_[0].shape[0]
        common_edge_counts = np.zeros((n_vars, n_vars))
        hidden_common_edge_counts = np.zeros((n_vars, n_vars))

        # Extract common edges from each bootstrap iteration
        for i in range(n_samples):
            nonzero_masks = []
            nan_masks = []
            for bootstrap_result in bootstrap_results:
                adj_matrix = bootstrap_result.adjacency_matrices_[i]

                # common edge: non-zero and not NaN
                valid_mask = ~np.isnan(adj_matrix)
                nonzero_mask = (adj_matrix != 0) & valid_mask
                nonzero_masks.append(nonzero_mask)

                # hidden common cause edges: NaN
                nan_mask = np.isnan(adj_matrix)
                nan_masks.append(nan_mask)

            common_nonzero_mask = np.logical_and.reduce(nonzero_masks)
            common_nan_mask = np.logical_and.reduce(nan_masks)
            common_edge_counts += common_nonzero_mask.astype(int)
            hidden_common_edge_counts += common_nan_mask.astype(int)

        # Combine probabilities
        common_edge_prob = common_edge_counts / n_samples
        hidden_common_edge_prob = hidden_common_edge_counts / n_samples

        # Stack into a single array of shape (2, n, n)
        common_edge_probabilities = np.stack((common_edge_prob, hidden_common_edge_prob), axis=0)

    else:
        raise ValueError("Invalid mode. Choose 'per' or 'across'.")

    return common_edge_probabilities


def print_common_edge_directions(common_edge_probabilities, confidence_level=0.01, labels=None):
    """
    Print statistically significant causal directions based on common edge probabilities.

    Parameters
    ----------
    common_edge_probabilities : ndarray
        A 2D NumPy array where each element [i, j] represents the probability
        that variable j causes variable i. The matrix is expected to be square
        and contain non-negative values.

    confidence_level : float, optional
        A threshold between 0 and 1 (inclusive) used to determine whether a causal
        direction is considered statistically significant. Only edges with probabilities
        strictly greater than this value will be displayed.

    labels : list of str, optional
        A list of variable names corresponding to the indices in the matrix.
        If provided, these labels will be used in the output instead of generic
        variable names (e.g., x0, x1, ...).
    """
    # check parameters
    if not 0.0 <= confidence_level < 1.0:
        raise ValueError("confidence_level must be a value between 0 and 1.")

    if not isinstance(common_edge_probabilities, np.ndarray):
        raise TypeError("common_edge_probabilities must be a NumPy ndarray.")

    if common_edge_probabilities.ndim != 3:
        raise ValueError("common_edge_probabilities must be a 3-dimensional array.")

    if common_edge_probabilities.shape[0] != 2:
        raise ValueError("The first dimension of common_edge_probabilities must be 2 (presence and unobserved probabilities).")

    n_rows, n_cols = common_edge_probabilities.shape[1], common_edge_probabilities.shape[2]

    # Probability matrix of shared edges
    joint_presence_prob = common_edge_probabilities[0]
    directions = []
    for from_ in range(n_cols):
        for to in range(n_rows):
            prob = joint_presence_prob[to, from_]
            if prob > confidence_level:
                directions.append((from_, to, prob, "<---"))

    # Probability matrix of NaNs due to shared unobserved common causes
    joint_unobserved_prob = common_edge_probabilities[1]
    bidirectional = set()
    for from_ in range(n_cols):
        for to in range(n_rows):
            if (to, from_) in bidirectional:
                continue
            prob = joint_unobserved_prob[to, from_]
            if prob > confidence_level:
                # In VAR-LiNGAM, due to mismatched matrix dimensions
                if from_ < n_rows:
                    reverse_prob = joint_unobserved_prob[from_, to]
                    if prob == reverse_prob:
                        bidirectional.add((to, from_))
                        bidirectional.add((from_, to))
                directions.append((from_, to, prob, "<-->"))

    directions.sort(key=lambda x: -x[2])

    for (from_, to, prob, edge_d) in directions:
        if labels:
            print(f"{labels[to]} {edge_d} {labels[from_]} ({100*prob:.1f}%)")
        else:
            print(f"x{to} {edge_d} x{from_} ({100*prob:.1f}%)")


def make_dot_for_nan_probability_matrix(
    adjacency_matrix,
    labels=None,
    lower_limit=0.01,
    path=None,
    path_color=None,
    ignore_shape=False,
):
    """Directed graph source code in the DOT language with specified nan probability matrix of hidden common cause.

    Parameters
    ----------
    adjacency_matrix : array-like with shape (n_features, n_features)
        Nan probability matrix of hidden common cause to make graph, where ``n_features`` is the number of features.
    labels : array-like, optional (default=None)
        Label to use for graph features.
    lower_limit : float, optional (default=0.01)
        Threshold for drawing direction.
        If float, then directions with absolute values of coefficients less than ``lower_limit`` are excluded.
    path : tuple, optional (default=None)
        Path to highlight. Tuple of start index and end index.
    path_color : string, optional (default=None)
        Colors to highlight a path.
    ignore_shape : boolean, optional (default=False)
        Ignore checking the shape of adjaceny_matrix or not.

    Returns
    -------
    graph : graphviz.Digraph
        Directed graph source code in the DOT language.
        If order is unknown, draw a double-headed arrow.
    """
    # Check parameters
    B = check_array(np.nan_to_num(adjacency_matrix))
    if not ignore_shape and B.shape[0] != B.shape[1]:
        raise ValueError("'adjacency_matrix' is not square matrix.")
    if labels is not None:
        if B.shape[1] != len(labels):
            raise ValueError(
                "Length of 'labels' does not match length of 'adjacency_matrix'"
            )

    if path is not None:
        if not isinstance(path, tuple) or len(path) < 2:
            raise TypeError("'path' should be a tuple of node indices.")
        if path_color is not None and not is_color_like(path_color):
            raise ValueError("'path_color' should be an color name.")

        if path_color is None:
            path_color = "black"
        path_start = check_scalar(
            path[0], "path_start", int, min_val=0, max_val=B.shape[0]
        )
        path_end = check_scalar(path[1], "path_end", int, min_val=0, max_val=B.shape[0])

    idx = np.abs(B) > lower_limit

    nx_graph = None
    if path is not None:
        nx_graph = nx.from_numpy_array(idx.astype(int).T, create_using=nx.DiGraph)

    # search path
    path_edges = None
    path_nodes = None
    if path is not None:
        path_edges = set()
        for path_ in nx.all_simple_paths(nx_graph, path[0], path[1]):
            path_edges |= set(zip(path_, path_[1:]))

        path_nodes = np.unique([list(edge) for edge in path_edges])
        if labels is None:
            path_nodes = [f"x{node}" for node in path_nodes]
        else:
            path_nodes = [labels[node] for node in path_nodes]

    d = graphviz.Digraph(engine="dot")

    # nodes
    names = labels if labels else [f"x{i}" for i in range(len(B))]
    for name in names:
        kwargs = {}
        if path is not None:
            if name in path_nodes:
                kwargs["fontcolor"] = path_color
                kwargs["color"] = path_color
            if name == names[path_start] or name == names[path_end]:
                kwargs["peripheries"] = "2"
        d.node(name, **kwargs)

    # edges
    # This matrix is a probability matrix of nan, and the diagonal components contain the same value
    dirs = np.where(np.tril(idx))
    for to, from_, coef in zip(dirs[0], dirs[1], B[dirs]):
        kwargs = {}
        if path is not None and (from_, to) in path_edges:
            kwargs["penwidth"] = "2"
            kwargs["fontcolor"] = path_color
            kwargs["color"] = path_color
        if from_ != to:
            if idx[to, from_] and B[from_, to] == B[to, from_]:
                d.edge(names[from_], names[to], dir="both", label=f"{coef:.2f}", **kwargs)

    return d



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import os

import numpy as np
import sys
sys.path.append('D:/Codes/Git/lingam')
import lingam

DATA_DIR_PATH = os.path.dirname(__file__)


def test_fit_lim():
    X = np.loadtxt(f"{DATA_DIR_PATH}/test_lim_data.csv", delimiter=",")
    dis_con = np.array([[0.0, 0.0, 0.0, 1.0, 1.0]])
    W_true = np.array(
        [
            [0.0, 1.09482609, -1.29270764, 0.0, -0.84424137],
            [0.0, 0.0, 0.80393307, 0.0, 0.0],
            [0.0, 0.0, 0.0, 0.0, 0.0],
            [0.70346053, 0.0, 1.90912441, 0.0, 1.94441713],
            [0.0, 0.0, -0.63152585, 0.0, 0.0],
        ]
    )
    model = lingam.LiM()
    model.fit(X, dis_con, only_global=False, is_poisson=True)

    print("The estimated adjacency matrix is:\n", model.adjacency_matrix_)
    print("The true adjacency matrix is:\n", W_true)
    print("Done.")

    model = lingam.LiM()
    model.fit(X, dis_con, only_global=False, is_poisson=False)
    # model.fit(X, dis_con, only_global=False)

    print("The estimated adjacency matrix is:\n", model.adjacency_matrix_)
    print("The true adjacency matrix is:\n", W_true)
    print("Done.")



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import copy
import itertools

import numpy as np
from pygam import LinearGAM
from scipy.stats import chi2
from sklearn.utils import check_array

from .hsic import hsic_test_gamma
from .utils import f_correlation


class MultiGroupCAMUV:
    """Implementation of CAM-UV Algorithm with multiple groups [1]_ [2]_ [3]_

    References
    ----------
    .. [1] T.N.Maeda and S.Shimizu. Causal additive models with unobserved variables.
       In Proc. Thirty-Seventh Conference on Uncertainty in Artificial Intelligence (UAI). PMLR  161:97-106, 2021.
    .. [2] T. N. Maeda and S. Shimizu. Use of prior knowledge to discover causal additive models with unobserved
       variables and its application to time series data. Behaviormetrika, xx(xx): 1-19, 2024.
    .. [3] S. Shimizu. Joint estimation of linear non-Gaussian acyclic models. Neurocomputing, 81: 104-107, 2012.
    """

    def __init__(
        self,
        alpha=0.01,
        num_explanatory_vals=2,
        independence="hsic",
        ind_corr=0.5,
        prior_knowledge=None,
    ):
        """Construct a CAM-UV model.

        Parameters
        ----------
        alpha : float, optional (default=0.01)
            Alpha level.
        num_explanatory_vals : int, optional (default=2)
            Maximum number of explanatory variables.
        independence : {'hsic', 'fcorr'}, optional (default='hsic')
            Methods to determine independence. If 'hsic' is set, test for independence by HSIC.
            If 'fcorr' is set, independence is determined by F-correlation.
        ind_corr : float, optional (default=0.5)
            The threshold value for determining independence by F-correlation;
            independence is determined when the value of F-correlation is below this threshold value.
        prior_knowledge : array-like, shape ((index, index), ...), optional (default=None)
            List of variable pairs indicating prior knowledge.
            If (0, 3) is included , it means that X0 cannot be a cause of X3.
        """

        # Check parameters
        if num_explanatory_vals <= 0:
            raise ValueError("num_explanatory_vals must be > 0.")

        if alpha < 0:
            raise ValueError("alpha must be >= 0.")

        if independence not in ("hsic", "fcorr"):
            raise ValueError("independence must be 'hsic' or 'fcorr'.")

        if ind_corr < 0.0:
            raise ValueError("ind_corr must be an float greater than 0.")

        self._num_explanatory_vals = num_explanatory_vals
        self._alpha = alpha
        self._independence = independence
        self._ind_corr = ind_corr
        self._pk_dict = self._make_pk_dict(prior_knowledge)

    def fit(self, X_list):
        """Fit the model to multiple datasets.

        Parameters
        ----------
        X_list : list, shape [X, ...]
            Multiple datasets for training, where ``X`` is an dataset.
            The shape of ''X'' is (n_samples, n_features),
            where ``n_samples`` is the number of samples and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """

        X_list = self._check_X_list(X_list)

        d = X_list[0].shape[1]
        N = self._get_neighborhoods(X_list)
        P = self._find_parents(X_list, self._num_explanatory_vals, N)

        U = []
        for i in range(d):
            for j in range(d)[i + 1 :]:
                if (i in P[j]) or (j in P[i]):
                    continue
                if (i not in N[j]) or (j not in N[i]):
                    continue

                in_X_list = []
                in_Y_list = []
                for X in X_list:
                    n = X.shape[0]
                    i_residual = self._get_residual(X, i, P[i])
                    j_residual = self._get_residual(X, j, P[j])
                    in_X_list.append(np.reshape(i_residual, [n, 1]))
                    in_Y_list.append(np.reshape(j_residual, [n, 1]))

                if not self._is_independent(in_X_list, in_Y_list):
                    if not set([i, j]) in U:
                        U.append(set([i, j]))

        self._U = U
        self._P = P

        return self._estimate_adjacency_matrix(X_list[0], P, U)

    def _make_pk_dict(self, prior_knowledge):
        if prior_knowledge is None:
            return None

        pk_dict = dict()
        for pair in prior_knowledge:
            if not pair[1] in pk_dict:
                pk_dict[pair[1]] = [pair[0]]
            else:
                pk_dict[pair[1]].append(pair[0])
        return pk_dict

    def _check_X_list(self, X_list):
        """Check input X list."""
        if not isinstance(X_list, list):
            raise ValueError("X_list must be a list.")

        if len(X_list) < 2:
            raise ValueError("X_list must be a list containing at least two items")

        self._k = len(X_list)
        self._n_features = check_array(X_list[0]).shape[1]
        X_list_ = []
        for X in X_list:
            X_ = check_array(X)
            if X_.shape[1] != self._n_features:
                raise ValueError(
                    "X_list must be a list with the same number of features"
                )
            X_list_.append(X_)

        return X_list_

    def _get_residual(self, X, explained_i, explanatory_ids):
        explanatory_ids = list(explanatory_ids)

        if len(explanatory_ids) == 0:
            residual = X[:, explained_i]
        else:
            gam = LinearGAM().fit(X[:, explanatory_ids], X[:, explained_i])
            residual = X[:, explained_i] - gam.predict(X[:, explanatory_ids])
        return residual

    def _is_independent(self, X_list, Y_list):
        if self._independence == "hsic":
            threshold = self._alpha
        elif self._independence == "fcorr":
            threshold = self._ind_corr
        is_independent, _ = self._is_independent_by(X_list, Y_list, threshold)
        return is_independent

    def _is_independent_by(self, X_list, Y_list, threshold):
        is_independent = False
        value = 0.0
        if self._independence == "hsic":
            fisher_stat = 0
            for X, Y in zip(X_list, Y_list):
                _, hsic_p = hsic_test_gamma(X, Y)
                fisher_stat += np.inf if hsic_p == 0 else -2 * np.log(hsic_p)
            fisher_p = chi2.sf(fisher_stat, df=2 * self._k)
            is_independent = fisher_p > threshold
            value = fisher_p
        elif self._independence == "fcorr":
            max_f_corr = 0.0
            for X, Y in zip(X_list, Y_list):
                f_corr = f_correlation(X, Y)
                if f_corr > max_f_corr:
                    max_f_corr = f_corr
            is_independent = max_f_corr < threshold
            value = max_f_corr
        return is_independent, value

    def _get_neighborhoods(self, X_list):
        d = X_list[0].shape[1]
        N = [set() for i in range(d)]
        for i in range(d):
            for j in range(d)[i + 1 :]:

                in_X_list = []
                in_Y_list = []
                for X in X_list:
                    n = X.shape[0]
                    in_X_list.append(np.reshape(X[:, i], [n, 1]))
                    in_Y_list.append(np.reshape(X[:, j], [n, 1]))

                if not self._is_independent(in_X_list, in_Y_list):
                    N[i].add(j)
                    N[j].add(i)
        return N

    def _find_parents(self, X_list, maxnum_vals, N):
        d = X_list[0].shape[1]
        P = [set() for i in range(d)]  # Parents
        t = 2
        Y_list = [copy.deepcopy(X) for X in X_list]

        while True:
            changed = False
            variables_set_list = list(itertools.combinations(set(range(d)), t))
            for variables_set in variables_set_list:
                variables_set = set(variables_set)

                if not self._check_identified_causality(variables_set, P):
                    continue

                child, is_independence_with_K = self._get_child(
                    X_list, variables_set, P, N, Y_list
                )
                if child is None:
                    continue
                if not is_independence_with_K:
                    continue

                parents = variables_set - {child}
                if not self._check_independence_withou_K(parents, child, P, N, Y_list):
                    continue

                for parent in parents:
                    P[child].add(parent)
                    changed = True
                    Y_list_ = []
                    for X, Y in zip(X_list, Y_list):
                        Y_list_.append(self._get_residuals_matrix(X, Y, P, child))
                    Y_list = Y_list_

            if changed:
                t = 2
            else:
                t += 1
                if t > maxnum_vals:
                    break

        for i in range(d):
            non_parents = set()
            for j in P[i]:

                in_X_list = []
                in_Y_list = []
                for X in X_list:
                    n = X.shape[0]
                    residual_i = self._get_residual(X, i, P[i] - {j})
                    residual_j = self._get_residual(X, j, P[j])
                    in_X_list.append(np.reshape(residual_i, [n, 1]))
                    in_Y_list.append(np.reshape(residual_j, [n, 1]))

                if self._is_independent(in_X_list, in_Y_list):
                    non_parents.add(j)
            P[i] = P[i] - non_parents

        return P

    def _check_prior_knowledge(self, xj_list, xi):
        if self._pk_dict is not None:
            for xj in xj_list:
                if (xi in self._pk_dict) and (xj in self._pk_dict[xi]):
                    return True
        return False

    def _get_residuals_matrix(self, X, Y_old, P, child):
        Y = copy.deepcopy(Y_old)
        Y[:, child] = self._get_residual(X, child, P[child])
        return Y

    def _get_child(self, X_list, variables_set, P, N, Y_list):
        prev_independence = 0.0 if self._independence == "hsic" else 1.0
        max_independence_child = None

        for child in variables_set:
            parents = variables_set - {child}

            if self._check_prior_knowledge(parents, child):
                continue

            if not self._check_correlation(child, parents, N):
                continue

            in_X_list = []
            in_Y_list = []
            for X, Y in zip(X_list, Y_list):
                n = X.shape[0]
                residual = self._get_residual(X, child, parents | P[child])
                in_X_list.append(np.reshape(residual, [n, 1]))
                in_Y_list.append(np.reshape(Y[:, list(parents)], [n, len(parents)]))

            is_ind, value = self._is_independent_by(
                in_X_list, in_Y_list, prev_independence
            )
            if is_ind:
                prev_independence = value
                max_independence_child = child

        if self._independence == "hsic":
            is_independent = prev_independence > self._alpha
        elif self._independence == "fcorr":
            is_independent = prev_independence < self._ind_corr

        return max_independence_child, is_independent

    def _check_independence_withou_K(self, parents, child, P, N, Y_list):
        for parent in parents:

            in_X_list = []
            in_Y_list = []
            for Y in Y_list:
                n = Y.shape[0]
                in_X_list.append(np.reshape(Y[:, child], [n, 1]))
                in_Y_list.append(np.reshape(Y[:, parent], [n, 1]))

            if self._is_independent(in_X_list, in_Y_list):
                return False
        return True

    def _check_identified_causality(self, variables_set, P):
        variables_list = list(variables_set)
        for i in variables_list:
            for j in variables_list[variables_list.index(i) + 1 :]:
                if (j in P[i]) or (i in P[j]):
                    return False
        return True

    def _check_correlation(self, child, parents, N):
        for parent in parents:
            if parent not in N[child]:
                return False
        return True

    def _estimate_adjacency_matrix(self, X, P, U):
        B = np.zeros([X.shape[1], X.shape[1]], dtype="float64")
        for i, parents in enumerate(P):
            for parent in parents:
                B[i, parent] = 1
        for confounded_pair in U:
            confounded_pair = list(confounded_pair)
            B[confounded_pair[0], confounded_pair[1]] = np.nan
            B[confounded_pair[1], confounded_pair[0]] = np.nan
        self._adjacency_matrix = B
        return self

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features, n_features)
            The adjacency matrix B of fitted model, where
            n_features is the number of features.
        """
        return self._adjacency_matrix



import os
import warnings

import numpy as np
import pandas as pd
from lingam import HighDimDirectLiNGAM


def test_fit_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3'])

    model = HighDimDirectLiNGAM()
    model.fit(X)

    # causal order
    co = model.causal_order_
    assert co.index(0) < co.index(1) < co.index(3)

    # J
    model = HighDimDirectLiNGAM(J=6)
    model.fit(X)
    co = model.causal_order_
    assert co.index(0) < co.index(1) < co.index(3)

    # K
    model = HighDimDirectLiNGAM(K=2)
    model.fit(X)
    co = model.causal_order_
    assert co.index(0) < co.index(1) < co.index(3)

    # alpha
    model = HighDimDirectLiNGAM(alpha=0.1)
    model.fit(X)
    co = model.causal_order_
    assert co.index(0) < co.index(1) < co.index(3)

    # estimate_adj_mat
    model = HighDimDirectLiNGAM(estimate_adj_mat=False)
    model.fit(X)
    co = model.causal_order_
    assert co.index(0) < co.index(1) < co.index(3)
    assert model.adjacency_matrix_ is None

    # random_state
    model = HighDimDirectLiNGAM(random_state=0)
    model.fit(X)
    model2 = HighDimDirectLiNGAM(random_state=0)
    model2.fit(X)
    assert np.isclose(model.adjacency_matrix_, model2.adjacency_matrix_).all()

    # n_samples <= n_features
    X2 = np.random.uniform(-1, 1, size=(10, 10))

    model = HighDimDirectLiNGAM()
    with warnings.catch_warnings(record=True) as w:
        model.fit(X2)
        assert w[0].category == UserWarning

def test_fit_invalid_data():
    # Not array data
    X = 1
    try:
        model = HighDimDirectLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=5)
    x1 = np.array(['X', 'Y', 'X', 'Y', 'X'])
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    try:
        model = HighDimDirectLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.nan
    try:
        model = HighDimDirectLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.inf
    try:
        model = HighDimDirectLiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # J > 2
    try:
        model = HighDimDirectLiNGAM(J=2)
    except ValueError:
        pass
    else:
        raise AssertionError

    # J must be integer
    try:
        model = HighDimDirectLiNGAM(J=4.0)
    except TypeError:
        pass
    else:
        raise AssertionError

    # K >= 1
    try:
        model = HighDimDirectLiNGAM(K=0)
    except ValueError:
        pass
    else:
        raise AssertionError

    # K must be integer
    try:
        model = HighDimDirectLiNGAM(K=2.0)
    except TypeError:
        pass
    else:
        raise AssertionError

    # 0 <= alpha <= 1
    try:
        model = HighDimDirectLiNGAM(alpha=-0.1)
    except ValueError:
        pass
    else:
        raise AssertionError



import numpy as np
import pandas as pd
from lingam.ica_lingam import ICALiNGAM


def test_fit_success():
    # causal direction: x0 --> x1 --> x3
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    x2 = np.random.uniform(size=1000)
    x3 = 4.0 * x1 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1, x2, x3]).T, columns=['x0', 'x1', 'x2', 'x3'])

    model = ICALiNGAM()
    model.fit(X)

    # check the causal ordering
    co = model.causal_order_
    assert co.index(0) < co.index(1) < co.index(3)

    # check the adjacency matrix
    am = model.adjacency_matrix_
    # assert am[1, 0] > 1.5 and am[3, 1] > 3.5

    am[1, 0] = 0.0
    am[3, 1] = 0.0
    # assert np.sum(am) < 0.1

    # for coverage
    matrix = np.array([
        [0, 1, 1, 1],
        [0, 0, 1, 1],
        [0, 0, 0, 1],
        [0, 0, 0, 0],
    ])
    model = ICALiNGAM()
    model._search_causal_order(matrix)

    # for coverage
    matrix = np.array([
        [1, 1, 1],
        [1, 1, 1],
        [0, 0, 0],
    ])
    model = ICALiNGAM()
    model._search_causal_order(matrix)


def test_fit_invalid_data():
    # Not array data
    X = 1
    try:
        model = ICALiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include non-numeric data
    x0 = np.random.uniform(size=5)
    x1 = np.array(['X', 'Y', 'X', 'Y', 'X'])
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    try:
        model = ICALiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include NaN values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.nan
    try:
        model = ICALiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError

    # Include infinite values
    x0 = np.random.uniform(size=1000)
    x1 = 2.0 * x0 + np.random.uniform(size=1000)
    X = pd.DataFrame(np.array([x0, x1]).T, columns=['x0', 'x1'])
    X.iloc[100, 0] = np.inf
    try:
        model = ICALiNGAM()
        model.fit(X)
    except ValueError:
        pass
    else:
        raise AssertionError



import numpy as np
import pandas as pd

from sklearn.utils import check_array
from sklearn.base import is_regressor, is_classifier

import matplotlib.pyplot as plt
from matplotlib.figure import Figure


def visualize_nonlinear_causal_effect(
    X,
    cd_result,
    estimator,
    cause_name,
    effect_name,
    cause_positions=None,
    percentile=None,
    fig=None,
    boxplot=False,
):
    """Visualize non-linear causal effect.

    Parameters
    ----------
    X : pandas.DataFrame, shape (n_samples, n_features)
        Training data used to obtain cd_result.
    cd_result : array-like with shape (n_features, n_features) or BootstrapResult
        Adjacency matrix or BootstrapResult. These are the results of a causal discovery.
    estimator : estimator object
        ``estimator`` used for non-linear regression.
        Regression with ``estimator`` using ``cause_name`` and covariates as explanatory
        variables and ``effect_name`` as objective variable.
        Those covariates are searched for in ``cd_result``.
    cause_name : str
        The name of the cause variable.
    effect_name : str
        The name of the effect variable.
    cause_positions : array-like, optional (default=None)
        List of positions from which causal effects are calculated.
        By default, ``cause_positions`` stores the position at which the value range of X is divided
        into 10 equal parts.
    percentile : array-like, optional (default=None)
        A tuple consisting of three percentile values. Each value must be greater
        than 0 and less than 100. By default, (95, 50, 5) is set.
    fig : plt.Figure, optional (default=None)
        If ``fig`` is given, draw a figure in ``fig``. If not given, plt.fig
        is prepared internally.
    boxplot : boolean, optional (default=False)
        If True, draw a box plot instead of a scatter plot for each ``cause_positions``.

    Return
    ------
    fig : plt.Figure
        Plotted figure.
    """

    # check arguments
    if not isinstance(X, pd.DataFrame):
        raise TypeError("X must be pandas.DataFrame.")

    if not _is_bootstrap_result(cd_result) and not _is_adjacency_matrix(
        cd_result, X.shape[1]
    ):
        raise TypeError(
            "cd_result must be lingam.bootstrap.BootstrapResult or array-like."
        )

    if cause_name not in X.columns:
        raise ValueError("cause_name is not exsit in X.columns.")

    if X[cause_name].dtype == "category":
        raise ValueError("cause_name must not be categorical.")

    if effect_name not in X.columns:
        raise ValueError("effect_name is not exsit in X.columns.")

    is_discrete = X[effect_name].dtype == "category"
    _check_estimator(estimator, is_discrete)

    if cause_positions is not None:
        cause_positions = check_array(
            cause_positions, dtype=None, ensure_2d=False
        ).flatten()
        if None in cause_positions:
            raise TypeError("cause_positions must not include None.")

    if percentile is not None:
        if not isinstance(percentile, list) or len(percentile) != 3:
            raise TypeError("percentile must be a list of 3 floats.")
        if not all(map(lambda x: 0 <= x and x <= 100, percentile)):
            raise TypeError(
                "Elements of percentile must be between 0 and 100 inclusive."
            )

    if fig is not None:
        if not isinstance(fig, Figure):
            raise TypeError("fig must be matplotlib.figure.Figure.")

    # set initial values
    if cause_positions is None:
        cause_positions = np.linspace(X[cause_name].min(), X[cause_name].max(), 10)

    if fig is None:
        fig = plt.figure()

    cause_index = list(X.columns).index(cause_name)
    effect_index = list(X.columns).index(effect_name)

    effect_classes = [None] if not is_discrete else np.unique(X[effect_name])

    if percentile is None:
        percentile = (95, 50, 5)

    # calculate causal effects
    if _is_bootstrap_result(cd_result):
        adj_matrices = cd_result.adjacency_matrices_
        index_list = cd_result.resampled_indices_
        if index_list is None:
            index_list = [X.index]
        disable_percentile = False
    else:
        adj_matrices = [cd_result]
        index_list = [X.index]
        disable_percentile = True

    effect_info = _calculate(
        effect_classes,
        adj_matrices,
        index_list,
        cause_index,
        effect_index,
        estimator,
        cause_positions,
        X,
        is_discrete,
    )

    # visualize
    fig = _draw(
        effect_info,
        cause_positions,
        percentile,
        fig,
        cause_name,
        effect_name,
        boxplot,
        disable_percentile,
    )

    return fig


def _is_bootstrap_result(cd_result):
    name = type(cd_result).__module__ + "." + type(cd_result).__name__
    if name != "lingam.bootstrap.BootstrapResult":
        return False
    return True


def _is_adjacency_matrix(cd_result, col_num):
    check_array(cd_result)

    if cd_result.shape[0] != col_num or cd_result.shape[1] != col_num:
        return False
    return True


def _check_estimator(estimator, is_discrete):
    if is_discrete is True and is_regressor(estimator):
        raise TypeError("X[effect_name] is discrete but the estimator is a regressor.")
    elif is_discrete is False and is_classifier(estimator):
        raise TypeError(
            "X[effect_name] is not discrete but the estimator is a classifier."
        )

    if is_classifier(estimator):
        try:
            func = getattr(estimator, "predict_proba")
            if not callable(func):
                raise Exception
        except Exception:
            raise RuntimeError("Classification models must have predict_proba().")


def _calculate(
    effect_classes,
    adj_matrices,
    index_list,
    cause_index,
    effect_index,
    estimator,
    cause_positions,
    X,
    is_discrete,
):
    # effect_classes is [None] if X.iloc[effect_index] is continuous
    effect_info = {effect_class: {} for effect_class in effect_classes}
    for c in effect_info.keys():
        effect_info[c] = {p: [] for p in cause_positions}

    for m, sub_index in zip(adj_matrices, index_list):
        # parents of cause except effect
        covariate = np.where(m[cause_index, :] == 1)[0].tolist()
        covariate = np.delete(covariate, np.argwhere(covariate == effect_index))

        # train
        exp = np.sort([cause_index] + covariate.tolist())
        exp = np.delete(exp, np.argwhere(exp == effect_index))
        X_exp = pd.get_dummies(X.iloc[sub_index, exp])
        X_obj = X.iloc[sub_index, effect_index]
        estimator.fit(X_exp, X_obj)

        # calculate causal effects
        for interv_val in cause_positions:
            # intervention
            X_ = X.iloc[sub_index, :].copy()
            X_.iloc[:, cause_index] = interv_val
            X_exp2 = pd.get_dummies(X_.iloc[:, exp])

            if not is_discrete:
                effect_values = estimator.predict(X_exp2)

                v = np.mean(effect_values)
                effect_info[None][interv_val].append(v)
            else:
                effect_values = estimator.predict_proba(X_exp2)

                for i, c in enumerate(estimator.classes_):
                    v = np.mean(effect_values[:, i])
                    effect_info[c][interv_val].append(v)

    return effect_info


def _draw(
    effect_info,
    cause_positions,
    percentile,
    fig,
    cause_name,
    effect_name,
    is_box,
    disable_percentile,
):
    if disable_percentile is False:
        alpha = 0.5
        colors = ["red", "green", "blue"]
    else:
        percentile = [50]
        alpha = 1.0
        colors = ["green"]

    for i in range(len(effect_info)):
        fig.add_subplot(1, len(effect_info), i + 1)

    axes = np.array(fig.get_axes()).flatten()

    for ax, (class_name, effects) in zip(axes, effect_info.items()):
        per_plot = {per: [] for per in percentile}

        for interv_val, values in effects.items():
            if not is_box:
                ax.scatter(
                    interv_val * np.ones(len(values)),
                    values,
                    color="black",
                    alpha=alpha,
                    facecolor="none",
                )
            else:
                ax.boxplot(values, positions=[interv_val], manage_ticks=False)

            per_points = np.percentile(values, percentile)
            for per_point, percentile_ in zip(per_points, percentile):
                per_plot[percentile_].append(per_point)

        ax.set_xlabel(cause_name)
        if class_name is None:
            ax.set_ylabel(f"causal effect {effect_name}")
        else:
            ax.set_ylim(-0.05, 1.05)
            ax.set_title(f"class {class_name}")
            ax.set_ylabel("probability")

        for (per, values), color in zip(per_plot.items(), colors):
            label = f"{per}%" if disable_percentile is False else None
            ax.plot(cause_positions, values, color=color, label=label)

    return fig



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""
import numbers
from abc import ABCMeta, abstractmethod

import numpy as np
from sklearn.utils import check_array, check_scalar, check_random_state
from sklearn.experimental import enable_iterative_imputer
from sklearn.impute import IterativeImputer

from ..multi_group_direct_lingam import MultiGroupDirectLiNGAM

__all__ = [
    "bootstrap_with_imputation",
    "BaseMultipleImputation",
    "BaseMultiGroupCDModel",
]


def bootstrap_with_imputation(
    X,
    n_sampling,
    n_repeats=10,
    imp=None,
    cd_model=None,
    prior_knowledge=None,
    apply_prior_knowledge_softly=False,
    random_state=None
):
    """Discovering causal relations in data with missing values..

    `bootstrap_with_imputation` is a function to perform a causal discovery
    on a dataset with missing values. `bootstrap_with_imputation` creates
    `n_sampling` bootstrap samples from the dataset, creates `n_repeats` samples
    for each bootstrap sample, completes the missing values in each sample,
    and runs a causal discovery assuming a common causal structure for
    `n_repeats` samples.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where ``n_samples`` is the number of samples
        and ``n_features`` is the number of features.
    n_sampling : int
        The number of bootstraps.
    n_repeats : int, optional (default=10)
        The number of times to complete missing values for each bootstrap sample.
        This value is only used when imp is None.
    imp : object, optional (default=None)
        Instance of a class inheriting from ``BaseMultipleImputation`` class.
        If None, this function uses ``_DefaultMultipleImputation`` to impute datasets.
    cd_model : object, optional (default=None)
        Instance of a class inheriting from ``BaseMultiGroupCDModel`` class.
        If None, this function uses ``MultiGroupDirectLiNGAM`` to estimate the causal order.
    prior_knowledge : array-like, shape (n_features, n_features), optional (default=None)
        Prior knowledge used for the causal discovery, where ``n_features`` is the number of features.
        prior_knowledge is used only if cd_model is None.

        The elements of prior knowledge matrix are defined as follows:

        * ``0`` : :math:`x_i` does not have a directed path to :math:`x_j`
        * ``1`` : :math:`x_i` has a directed path to :math:`x_j`
        * ``-1`` : No prior knowledge is available to know if either of the two cases above (0 or 1) is true.
    apply_prior_knowledge_softly : boolean, optional (default=False)
        If True, apply prior knowledge softly.
        ``apply_prior_knowledge_softly`` is used only if ``cd_model`` is None.
    random_state : int, optional (default=None)
        ``random_state`` is the seed used by the random number generator.

    Returns
    -------
    causal_orders : array-like, shape (n_sampling, n_features)
        The causal order of the fitted model, where
        n_features is the number of features.
    adj_matrices_list : array-like, shape (n_sampling, n_repeats, n_features, n_features)
        The list of adjacency matrices.
    resampled_indices_ : array-like, shape (n_sampling, n_samples)
        The list of original index of resampled samples.
    imputation_results : array-like, shape (n_sampling, n_repeats, n_samples, n_features)
        This array shows the result of the imputation.
        Elements which are not NaN are the imputation values.
    """
    # check args
    X = check_array(X, force_all_finite='allow-nan')

    n_sampling = check_scalar(
        n_sampling,
        "n_sampling",
        (numbers.Integral, np.integer),
        min_val=1
    )

    n_repeats = check_scalar(
        n_repeats,
        "n_repeats",
        (numbers.Integral, np.integer),
        min_val=1
    )

    if cd_model is not None and not isinstance(cd_model, BaseMultiGroupCDModel):
        raise ValueError("cd_model must be an instance of a subclass of BaseMultiGroupCDModel.")

    if imp is not None and not isinstance(imp, BaseMultipleImputation):
        raise ValueError("imp must be an instance of a subclass of BaseMultipleImputation.")

    n_samples, n_features = X.shape
    if prior_knowledge is not None:
        prior_knowledge = check_array(prior_knowledge)
        if prior_knowledge.shape != (n_features, n_features):
            raise ValueError("The shape of prior_knowledge must be (n_features, n_features).")

    random_state = check_random_state(random_state)

    if imp is None:
        imp = _DefaultMultipleImputation(n_repeats, random_state)

    if cd_model is None:
        cd_model = _DefaultMultiGroupCDModel(
            prior_knowledge=prior_knowledge,
            apply_prior_knowledge_softly=apply_prior_knowledge_softly,
            random_state=random_state
        )

    resampled_indices = []
    causal_orders = []
    adj_matrices_list = []
    imputation_results = []

    for i in range(n_sampling):
        # make a bootstrap sample
        resampled_index = random_state.choice(np.arange(X.shape[0]), replace=True, size=len(X))
        bootstrap_sample = X[resampled_index]

        # send bootstrap_sample that has not been imputed yet
        cd_model.before_imputation(bootstrap_sample)

        # make datasets
        datasets = imp.fit_transform(bootstrap_sample)
        datasets = _check_imputer_outout(datasets, n_samples, n_features)

        n_repeats_impl = len(datasets)

        # run causal discovery assuming a common causal structure
        result = cd_model.fit(datasets)
        causal_order, adjacency_matrices = _check_cd_output(result, n_repeats_impl, n_features)

        # store imputation results
        # hold values only if NaN
        datasets = np.array(datasets)
        imputation_result = np.full(datasets.shape, np.nan)
        pos = np.isnan(bootstrap_sample)
        imputation_result[:, pos] = datasets[:, pos]

        resampled_indices.append(resampled_index)
        causal_orders.append(causal_order)
        adj_matrices_list.append(adjacency_matrices)
        imputation_results.append(imputation_result)

    resampled_indices = np.array(resampled_indices)
    causal_orders = np.array(causal_orders)
    adj_matrices_list = np.array(adj_matrices_list)
    imputation_results = np.array(imputation_results)

    return causal_orders, adj_matrices_list, resampled_indices, imputation_results


class BaseMultipleImputation(metaclass=ABCMeta):
    """ The abstract class of the causal discovery model for the multigroup data

    Inherit this abstract class and send that instance to ``bootstrap_with_imputation``
    if you need to customize the causal discovery model in ``bootstrap_with_imputation``.
    """

    @abstractmethod
    def fit_transform(self, X):
        """
        This method is called to fit imputed bootstrap sample.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Target data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
            ``X`` may contain missing values.

        Returns
        -------
        X_list: list, shape [X, ...]
            The list of imputed X.
        """
        raise NotImplementedError


def _check_imputer_outout(imp_output, n_samples, n_features):
    try:
        imputed_data = check_array(imp_output, allow_nd=True)
    except Exception as e:
        raise ValueError("The return value of imp violates its specification: " + str(e))
    
    if imputed_data.shape[1:] != (n_samples, n_features):
        raise ValueError("The shape of the return value of imp must be (n_repeats, n_samples, n_fatures).")

    imputed_data = list(imputed_data)

    return imputed_data


class BaseMultiGroupCDModel(metaclass=ABCMeta):
    """ The abstract class of the causal discovery model for the multigroup data

    Inherit this abstract class and send that instance to ``bootstrap_with_imputation``
    if you need to customize the causal discovery model in ``bootstrap_with_imputation``.
    """

    @abstractmethod
    def before_imputation(self, X):
        """
        This method is called just before the bootstrap sample is imputed.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            ``X`` is a bootstrap sample and has not yet been imputed.
            ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
            ``X`` may contain missing values.

        Returns
        -------
        None
        """
        raise NotImplementedError

    @abstractmethod
    def fit(self, X_list):
        """
        This method is called to fit imputed bootstrap sample.

        Parameters
        ----------
        X_list : list, shape [X, ...]
            ``X_list`` is a list of the imputed bootstrap samples.
            Multiple datasets for training, where ``X`` is an dataset.
            The shape of ''X'' is (n_samples, n_features),
            where ``n_samples`` is the number of samples and ``n_features`` is the number of features.
            Each ``X`` may contain missing values.

        Returns
        -------
        causal_order : array-like
            The estimated causal order.
        adjacenyc_matrices : array-like
            The estimated adjacency matrices.
        """
        raise NotImplementedError


def _check_cd_output(cd_output, n_repeats, n_features):
    if not isinstance(cd_output, tuple) or len(cd_output) != 2:
        raise ValueError("The return value of cd_model.fit() must be a tuple like (causal_order, adjacenecy_matrices).")

    causal_order, adjacency_matrices = cd_output

    try:
        causal_order = check_array(causal_order, ensure_2d=False)
    except Exception as e:
        raise ValueError("causal_order, the output of cd_model, violates its specification: " + str(e))

    if len(causal_order) != n_features:
        raise ValueError("The length of causal_order, the output of cd_model, must be equal to n_features.")

    if not np.array_equal(np.unique(causal_order), np.arange(len(causal_order))):
        raise ValueError("Elements of causal_order, the output of cd_model, must be unique and must be indicates a column number.")

    try:
        adjacency_matrices = check_array(adjacency_matrices, ensure_all_finite="allow-nan", allow_nd=True)
    except Exception as e:
        raise ValueError("adjacency_matrices, the output of cd_model, violates its specification: " + str(e))

    if adjacency_matrices.shape[-2:] != (n_features, n_features):
        raise ValueError("The shape of elements of adjacency_matrices, the output of cd_model, must be (n_features, n_features)")

    return causal_order, adjacency_matrices


class _DefaultMultipleImputation(metaclass=ABCMeta):
    """ The default class for the multiple imputation """

    def __init__(self, n_repeats, random_state):
        self._imp = IterativeImputer(sample_posterior=True, random_state=random_state)
        self._n_repeats = n_repeats

    def fit_transform(self, X):
        X_list = []
        for i in range(self._n_repeats):
            X_ = self._imp.fit_transform(X)
            X_list.append(X_)

        return X_list


class _DefaultMultiGroupCDModel(BaseMultiGroupCDModel):
    """ The default class for the causal discovery on the multigroup data """

    def __init__(self, prior_knowledge=None, apply_prior_knowledge_softly=False, random_state=None):
        self._model = MultiGroupDirectLiNGAM(
            prior_knowledge=prior_knowledge,
            apply_prior_knowledge_softly=apply_prior_knowledge_softly,
            random_state=random_state
        )

    def before_imputation(self, X):
        pass

    def fit(self, X_list):
        self._model.fit(X_list)
        return self._model.causal_order_, self._model.adjacency_matrices_



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/view/sshimizu06/lingam
"""

import itertools
import numbers

import numpy as np
from scipy.stats.distributions import chi2
from sklearn.linear_model import LinearRegression
from sklearn.utils import check_array, resample

from .utils import predict_adaptive_lasso, calculate_total_effect
from .hsic import hsic_test_gamma
from .bootstrap import BootstrapResult


class GroupLiNGAM:
    """Implementation of GroupLiNGAM Algorithm [1]_

    References
    ----------
    .. [1] Y. Kawahara, K. Bollen, S. Shimizu and T. Washio.
       GroupLiNGAM: Linear non-Gaussian acyclic models for sets of variables.
       Arxiv preprint arXiv:1006.5041, 2010.
    """
    def __init__(self, alpha=0.01):
        """Construct a GroupLiNGAM model.

        Parameters
        ----------
        alpha : float, optional (default=0.01)
            Alpha level for HSIC independence test.
        """
        # Check parameters
        if alpha < 0:
            raise ValueError("alpha must be >= 0.")

        self._alpha = alpha
        self._adjacency_matrix = None
        self._causal_order = []

    def fit(self, X):
        """Fit the model to X.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        # Check parameters
        X = check_array(X)
        n_features = X.shape[1]

        # Causal discovery
        V = list(np.arange(n_features))
        self._causal_order = self._group_search(V, X)

        return self._estimate_adjacency_matrix(X)

    def _compute_residuals(self, predictors, target):
        """Compute residuals"""
        lr = LinearRegression().fit(predictors, target)
        resid = target - lr.predict(predictors)
        return resid

    def _fisher_hsic_test(self, X, Y):
        """Conduct statistical test by HSIC and Fisher's method for multivariate X and Y."""
        fisher_stat = 0
        n_features_X = X.shape[1]
        n_features_Y = Y.shape[1]

        for i in range(n_features_X):
            for j in range(n_features_Y):
                _, hsic_p = hsic_test_gamma(X[:, [i]], Y[:, [j]])
                fisher_stat += np.inf if hsic_p == 0 else -2 * np.log(hsic_p)

        df = 2 * n_features_X * n_features_Y
        fisher_p = chi2.sf(fisher_stat, df=df)

        return fisher_stat, fisher_p

    def _group_search(self, U, X_U):
        """Recursively devide the given subset U into independent groups"""
        # Generate the subsets
        subsets = []
        if len(U) == 1:
            return [U]
        else:
            for num in range(1, len(U)):
                subsets.extend([list(comb) for comb in itertools.combinations(U, num)])

        hsic_p_list = []
        for S in subsets:
            # Compute the residuals
            U_S = np.setdiff1d(U, S)
            rS = self._compute_residuals(X_U[:, S], X_U[:, U_S])

            # Compute some independence
            _, hsic_p = self._fisher_hsic_test(X_U[:, S], rS)
            hsic_p_list.append(hsic_p)

        S_star_idx = np.argmax(hsic_p_list)
        S_star = subsets[S_star_idx]
        hsic_p_S_star = hsic_p_list[S_star_idx]

        K = []
        if hsic_p_S_star > self._alpha and len(U) > 1:
            # Compute the residuals
            U_S_star = np.setdiff1d(U, S_star)
            rS_star = self._compute_residuals(X_U[:, S_star], X_U[:, U_S_star])

            # Call the GroupSearch function recursively
            index_map = {val: i for i, val in enumerate(S_star)}
            local_S_star = [index_map[i] for i in U if i in index_map]
            K_S = self._group_search(local_S_star, X_U[:, S_star])
            K.extend([[S_star[i] for i in group] for group in K_S])

            index_map = {val: i for i, val in enumerate(U_S_star)}
            local_U_S = [index_map[i] for i in U if i in index_map]
            K_U_S = self._group_search(local_U_S, rS_star)
            K.extend([[U_S_star[i] for i in group] for group in K_U_S])
        else:
            K.append(U)

        return K

    def _estimate_adjacency_matrix(self, X):
        """Estimate adjacency matrix by causal order.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where n_samples is the number of samples
            and n_features is the number of features.

        Returns
        -------
        self : object
            Returns the instance itself.
        """
        B = np.zeros((X.shape[1], X.shape[1]), dtype="float64")
        for i in range(1, len(self._causal_order)):
            target_group = self._causal_order[i]
            predictor_groups = self._causal_order[:i]
            predictors = [var for group in predictor_groups for var in group]

            for target in target_group:
                B[target, predictors] = predict_adaptive_lasso(X, predictors, target)

        for target_group in self._causal_order:
            if len(target_group) > 1:
                for k, l in list(itertools.combinations(target_group, 2)):
                    B[k][l] = np.nan
                    B[l][k] = np.nan

        self._adjacency_matrix = B
        return self

    @property
    def adjacency_matrix_(self):
        """Estimated adjacency matrix.

        Returns
        -------
        adjacency_matrix_ : array-like, shape (n_features, n_features)
            The adjacency matrix B of fitted model, where
            n_features is the number of features.
        """
        return self._adjacency_matrix

    @property
    def causal_order_(self):
        """Estimated causal ordering.

        Returns
        -------
        causal_order_ : array-like, shape (n_features)
            The causal order of fitted model, where
            n_features is the number of features.
        """
        return self._causal_order

    def bootstrap(self, X, n_sampling):
        """Evaluate the statistical reliability of DAG based on the bootstrapping.

        Parameters
        ----------
        X : array-like, shape (n_samples, n_features)
            Training data, where ``n_samples`` is the number of samples
            and ``n_features`` is the number of features.
        n_sampling : int
            Number of bootstrapping samples.

        Returns
        -------
        result : BootstrapResult
            Returns the result of bootstrapping.
        """
        # Check parameters
        X = check_array(X)

        if isinstance(n_sampling, (numbers.Integral, np.integer)):
            if not 0 < n_sampling:
                raise ValueError("n_sampling must be an integer greater than 0.")
        else:
            raise ValueError("n_sampling must be an integer greater than 0.")

        # Bootstrapping
        adjacency_matrices = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        total_effects = np.zeros([n_sampling, X.shape[1], X.shape[1]])
        index = np.arange(X.shape[0])
        resampled_indices = []
        for i in range(n_sampling):
            resampled_X, resampled_index = resample(X, index)
            self.fit(resampled_X)
            adjacency_matrices[i] = self._adjacency_matrix

            # Calculate total effects
            for c, from_group in enumerate(self._causal_order):
                for to_group in self._causal_order[c + 1 :]:
                    for from_ in from_group:
                        for to in to_group:
                            total_effects[i, to, from_] = calculate_total_effect(
                                self._adjacency_matrix, from_, to
                            )

            resampled_indices.append(resampled_index)

        return BootstrapResult(adjacency_matrices, total_effects, resampled_indices=resampled_indices)
